@@ -83,6 +83,8 @@ template <> struct string_converter<Property>
 
 } // namespace PythonConversion
 
+std::string loc_repr_py(Loc loc) { return stringf("Loc(%d, %d, %d)", loc.x, loc.y, loc.z); }
+
 PYBIND11_EMBEDDED_MODULE(MODULE_NAME, m)
 {
     py::register_exception_translator([](std::exception_ptr p) {
@@ -175,7 +177,8 @@ PYBIND11_EMBEDDED_MODULE(MODULE_NAME, m)
                            .def(py::init<int, int, int>())
                            .def_readwrite("x", &Loc::x)
                            .def_readwrite("y", &Loc::y)
-                           .def_readwrite("z", &Loc::z);
+                           .def_readwrite("z", &Loc::z)
+                           .def("__repr__", loc_repr_py);
 
     auto ci_cls = py::class_<ContextualWrapper<CellInfo &>>(m, "CellInfo");
     readwrite_wrapper<CellInfo &, decltype(&CellInfo::name), &CellInfo::name, conv_to_str<IdString>,
@@ -217,7 +220,7 @@ PYBIND11_EMBEDDED_MODULE(MODULE_NAME, m)
     readwrite_wrapper<PortInfo &, decltype(&PortInfo::type), &PortInfo::type, pass_through<PortType>,
                       pass_through<PortType>>::def_wrap(pi_cls, "type");
 
-    typedef std::vector<PortRef> PortRefVector;
+    typedef indexed_store<PortRef> PortRefVector;
     typedef dict<WireId, PipMap> WireMap;
     typedef pool<BelId> BelSet;
     typedef pool<WireId> WireSet;
@@ -285,7 +288,7 @@ PYBIND11_EMBEDDED_MODULE(MODULE_NAME, m)
     WRAP_MAP(m, WireMap, wrap_context<PipMap &>, "WireMap");
     WRAP_MAP_UPTR(m, RegionMap, "RegionMap");
 
-    WRAP_VECTOR(m, PortRefVector, wrap_context<PortRef>);
+    WRAP_INDEXSTORE(m, PortRefVector, wrap_context<PortRef>);
 
     typedef dict<IdString, ClockFmax> ClockFmaxMap;
     WRAP_MAP(m, ClockFmaxMap, pass_through<ClockFmax>, "ClockFmaxMap");
