@@ -224,10 +224,14 @@ Arch::Arch(ArchArgs args) : args(args), disallow_site_routing(false)
 
     // Initially LutElement vectors for each tile type.
     tile_type_index = 0;
+    max_lut_cells = 0;
+    max_lut_pins = 0;
     lut_elements.resize(chip_info->tile_types.size());
     for (const TileTypeInfoPOD &tile_type : chip_info->tile_types) {
         std::vector<LutElement> &elements = lut_elements[tile_type_index++];
         elements.reserve(tile_type.lut_elements.size());
+
+        int lut_cells_count = 0;
         for (auto &lut_element : tile_type.lut_elements) {
             elements.emplace_back();
 
@@ -252,10 +256,15 @@ Arch::Arch(ArchArgs args) : args(args), disallow_site_routing(false)
                 }
 
                 lut.output_pin = IdString(lut_bel.out_pin);
+                lut_cells_count++;
+
+                max_lut_pins = std::max((int)lut_bel.pins.size(), max_lut_pins);
             }
 
             element.compute_pin_order();
         }
+
+        max_lut_cells = std::max(lut_cells_count, max_lut_cells);
     }
 
     // Map lut cell types to their LutCellPOD
@@ -1000,14 +1009,16 @@ delay_t Arch::estimateDelay(WireId src, WireId dst) const
 #endif
 }
 
-delay_t Arch::predictDelay(const NetInfo *net_info, const PortRef &sink) const
+delay_t Arch::predictDelay(BelId src_bel, IdString src_pin, BelId dst_bel, IdString dst_pin) const
 {
     // FIXME: Implement when adding timing-driven place and route.
+    NPNR_UNUSED(src_pin);
+    NPNR_UNUSED(dst_pin);
     int src_x, src_y;
-    get_tile_x_y(net_info->driver.cell->bel.tile, &src_x, &src_y);
+    get_tile_x_y(src_bel.tile, &src_x, &src_y);
 
     int dst_x, dst_y;
-    get_tile_x_y(sink.cell->bel.tile, &dst_x, &dst_y);
+    get_tile_x_y(dst_bel.tile, &dst_x, &dst_y);
 
     delay_t base = 30 * std::min(std::abs(dst_x - src_x), 18) + 10 * std::max(std::abs(dst_x - src_x) - 18, 0) +
                    60 * std::min(std::abs(dst_y - src_y), 6) + 20 * std::max(std::abs(dst_y - src_y) - 6, 0) + 300;
@@ -1375,7 +1386,7 @@ void Arch::merge_constant_nets()
             }
 
             NPNR_ASSERT(net->driver.port == gnd_cell_port);
-            std::vector<PortRef> users_copy = net->users;
+            indexed_store<PortRef> users_copy = net->users;
             for (const PortRef &port_ref : users_copy) {
                 IdString cell = port_ref.cell->name;
                 disconnectPort(cell, port_ref.port);
@@ -1398,7 +1409,7 @@ void Arch::merge_constant_nets()
             }
 
             NPNR_ASSERT(net->driver.port == vcc_cell_port);
-            std::vector<PortRef> users_copy = net->users;
+            indexed_store<PortRef> users_copy = net->users;
             for (const PortRef &port_ref : users_copy) {
                 IdString cell = port_ref.cell->name;
                 disconnectPort(cell, port_ref.port);
