@@ -700,7 +700,7 @@ struct Arch : ArchAPI<ArchRanges>
 
     // -------------------------------------------------
     delay_t estimateDelay(WireId src, WireId dst) const final;
-    delay_t predictDelay(const NetInfo *net_info, const PortRef &sink) const final;
+    delay_t predictDelay(BelId src_bel, IdString src_pin, BelId dst_bel, IdString dst_pin) const final;
     ArcBounds getRouteBoundingBox(WireId src, WireId dst) const final;
     delay_t getDelayEpsilon() const final { return 20; }
     delay_t getRipupDelayPenalty() const final { return 120; }
@@ -1135,6 +1135,11 @@ struct Arch : ArchAPI<ArchRanges>
     std::vector<std::vector<LutElement>> lut_elements;
     dict<IdString, const LutCellPOD *> lut_cells;
 
+    // Defines the max number of LUT cells in a site and LUT pins
+    // to allow a correct functioning of the site lut mapping cache
+    int max_lut_cells;
+    int max_lut_pins;
+
     // Of the LUT cells, which is used for wires?
     // Note: May be null in arch's without wire LUT types.  Assumption is
     // that these arch's don't need wire LUT's because the LUT share is simple
