@@ -23,6 +23,7 @@
 #include <math.h>
 #include <regex>
 #include "embed.h"
+#include "gfx.h"
 #include "nextpnr.h"
 #include "placer1.h"
 #include "placer_heap.h"
@@ -32,6 +33,148 @@
 
 NEXTPNR_NAMESPACE_BEGIN
 
+// GUI
+void Arch::fixClockSpineDecals(void)
+{
+    for (auto sp : clockSpinesCache) {
+        // row: (#of spine cells, wire_id)
+        dict<int, std::pair<int, IdString>> rows;
+        IdString min_x, max_x;
+        min_x = max_x = *sp.second.begin();
+        for (auto wire : sp.second) {
+            WireInfo &wi = wire_info(wire);
+            std::pair<int, IdString> &row = rows[wi.y];
+            ++row.first;
+            row.second = wire;
+            if (wi.x < wire_info(min_x).x) {
+                min_x = wire;
+            } else {
+                if (wi.x > wire_info(max_x).x) {
+                    max_x = wire;
+                }
+            }
+        }
+        // central mux row owns the global decal
+        int mux_row = -1;
+        for (auto row : rows) {
+            if (row.second.first == 1) {
+                mux_row = row.first;
+                break;
+            }
+        }
+        // if there is no separate central mux than all decals are the same
+        if (mux_row == -1) {
+            mux_row = rows.begin()->first;
+            WireInfo &wi = wire_info(rows.at(mux_row).second);
+            GraphicElement &el_active = decal_graphics.at(wi.decalxy_active.decal).at(0);
+            GraphicElement &el_inactive = decal_graphics.at(wi.decalxy_inactive.decal).at(0);
+            el_active.y1 -= wi.y;
+            el_active.y2 -= wi.y;
+            el_inactive.y1 -= wi.y;
+            el_inactive.y2 -= wi.y;
+            el_active.x1 += wire_info(min_x).x;
+            el_active.x2 += wire_info(max_x).x;
+            el_inactive.x1 += wire_info(min_x).x;
+            el_inactive.x2 += wire_info(max_x).x;
+        } else {
+            // change the global decal
+            WireInfo &wi = wire_info(rows.at(mux_row).second);
+            // clear spine decals
+            float y = 0.;
+            for (auto wire : sp.second) {
+                if (wire == wi.name) {
+                    continue;
+                }
+                wire_info(wire).decalxy_active = DecalXY();
+                wire_info(wire).decalxy_inactive = DecalXY();
+                y = wire_info(wire).y;
+            }
+            GraphicElement &el_active = decal_graphics.at(wi.decalxy_active.decal).at(0);
+            GraphicElement &el_inactive = decal_graphics.at(wi.decalxy_inactive.decal).at(0);
+            el_active.y1 -= y;
+            el_active.y2 -= y;
+            el_inactive.y1 -= y;
+            el_inactive.y2 -= y;
+            el_active.x1 += wire_info(min_x).x;
+            el_active.x2 += wire_info(max_x).x;
+            el_inactive.x1 += wire_info(min_x).x;
+            el_inactive.x2 += wire_info(max_x).x;
+        }
+        refreshUi();
+    }
+}
+
+void Arch::updateClockSpinesCache(IdString spine_id, IdString wire_id)
+{
+    std::vector<IdString> &sp = clockSpinesCache[spine_id];
+    if (std::find(sp.begin(), sp.end(), wire_id) == sp.end()) {
+        sp.push_back(wire_id);
+    }
+}
+
+DecalXY Arch::getBelDecal(BelId bel) const
+{
+    CellInfo *ci = getBoundBelCell(bel);
+    if (ci == nullptr) {
+        return bels.at(bel).decalxy_inactive;
+    } else {
+        // LUT + used/unused DFF
+        if (bels.at(bel).type == id_SLICE) {
+            DecalXY decalxy = bels.at(bel).decalxy_active;
+            if (!ci->params.at(id_FF_USED).as_bool()) {
+                decalxy.decal = id_DECAL_LUT_UNUSED_DFF_ACTIVE;
+                if (ci->params.count(id_ALU_MODE) != 0) {
+                    decalxy.decal = id_DECAL_ALU_ACTIVE;
+                }
+            }
+            return decalxy;
+        }
+    }
+    return bels.at(bel).decalxy_active;
+}
+
+DecalXY Arch::getGroupDecal(GroupId grp) const { return groups.at(grp).decalxy; }
+
+DecalXY Arch::getPipDecal(PipId pip) const
+{
+    if (getBoundPipNet(pip) == nullptr) {
+        return pips.at(pip).decalxy_inactive;
+    }
+    return pips.at(pip).decalxy_active;
+}
+
+DecalXY Arch::getWireDecal(WireId wire) const
+{
+    static std::vector<IdString> clk_wires = {id_GB00, id_GB10, id_GB20, id_GB30, id_GB40, id_GB50, id_GB60, id_GB70};
+    static std::vector<IdString> pip_dst = {id_CLK0, id_CLK1, id_CLK2, id_EW10, id_EW20, id_SN10, id_SN20};
+    if (getBoundWireNet(wire) == nullptr) {
+        if (std::find(clk_wires.begin(), clk_wires.end(), wires.at(wire).type) != clk_wires.end()) {
+            for (auto dst : pip_dst) {
+                // check if pip is used
+                char pip_name[20];
+                snprintf(pip_name, sizeof(pip_name), "%s_%s", wire.c_str(this), dst.c_str(this));
+                if (pips.count(id(pip_name)) != 0) {
+                    if (getBoundPipNet(id(pip_name)) != nullptr) {
+                        return wires.at(wire).decalxy_active;
+                    }
+                }
+            }
+        } else {
+            // spines
+            if (clockSpinesCache.count(wires.at(wire).type) != 0) {
+                std::vector<IdString> const &sp = clockSpinesCache.at(wires.at(wire).type);
+                for (auto w : sp) {
+                    if (getBoundWireNet(w) != nullptr) {
+                        return wires.at(wire).decalxy_active;
+                    }
+                }
+            }
+        }
+        return wires.at(wire).decalxy_inactive;
+    }
+    return wires.at(wire).decalxy_active;
+}
+
 WireInfo &Arch::wire_info(IdString wire)
 {
     auto w = wires.find(wire);
@@ -58,7 +201,6 @@ BelInfo &Arch::bel_info(IdString bel)
 
 void Arch::addWire(IdString name, IdString type, int x, int y)
 {
-    // std::cout << name.str(this) << std::endl;
     NPNR_ASSERT(wires.count(name) == 0);
     WireInfo &wi = wires[name];
     wi.name = name;
@@ -67,6 +209,16 @@ void Arch::addWire(IdString name, IdString type, int x, int y)
     wi.y = y;
 
     wire_ids.push_back(name);
+
+    // Needed to ensure empty tile bel locations
+    if (int(bels_by_tile.size()) <= x)
+        bels_by_tile.resize(x + 1);
+    if (int(bels_by_tile[x].size()) <= y)
+        bels_by_tile[x].resize(y + 1);
+    if (int(tileBelDimZ.size()) <= x)
+        tileBelDimZ.resize(x + 1);
+    if (int(tileBelDimZ[x].size()) <= y)
+        tileBelDimZ[x].resize(y + 1);
 }
 
 void Arch::addPip(IdString name, IdString type, IdString srcWire, IdString dstWire, DelayQuad delay, Loc loc)
@@ -105,6 +257,13 @@ void Arch::addPip(IdString name, IdString type, IdString srcWire, IdString dstWi
     tilePipDimZ[loc.x][loc.y] = std::max(tilePipDimZ[loc.x][loc.y], loc.z + 1);
 }
 
+void Arch::addGroup(IdString name)
+{
+    NPNR_ASSERT(groups.count(name) == 0);
+    GroupInfo &gi = groups[name];
+    gi.name = name;
+}
+
 void Arch::addBel(IdString name, IdString type, Loc loc, bool gb)
 {
     NPNR_ASSERT(bels.count(name) == 0);
@@ -189,24 +348,41 @@ void Arch::addDecalGraphic(DecalId decal, const GraphicElement &graphic)
     refreshUi();
 }
 
-void Arch::setWireDecal(WireId wire, DecalXY decalxy)
+void Arch::setWireDecal(WireId wire, DecalXY active, DecalXY inactive)
 {
-    wire_info(wire).decalxy = decalxy;
+    wire_info(wire).decalxy_active = active;
+    wire_info(wire).decalxy_inactive = inactive;
     refreshUiWire(wire);
 }
 
-void Arch::setPipDecal(PipId pip, DecalXY decalxy)
+void Arch::setPipDecal(PipId pip, DecalXY active, DecalXY inactive)
 {
-    pip_info(pip).decalxy = decalxy;
+    pip_info(pip).decalxy_active = active;
+    pip_info(pip).decalxy_inactive = inactive;
     refreshUiPip(pip);
 }
 
-void Arch::setBelDecal(BelId bel, DecalXY decalxy)
+void Arch::setBelDecal(BelId bel, DecalXY active, DecalXY inactive)
 {
-    bel_info(bel).decalxy = decalxy;
+    bel_info(bel).decalxy_active = active;
+    bel_info(bel).decalxy_inactive = inactive;
     refreshUiBel(bel);
 }
 
+void Arch::setDefaultDecals(void)
+{
+    for (BelId bel : getBels()) {
+        gfxSetBelDefaultDecal(this, bel_info(bel));
+    }
+    for (PipId pip : getPips()) {
+        gfxSetPipDefaultDecal(this, pip_info(pip));
+    }
+    for (WireId wire : getWires()) {
+        gfxSetWireDefaultDecal(this, wire_info(wire));
+    }
+    fixClockSpineDecals();
+}
+
 void Arch::setGroupDecal(GroupId group, DecalXY decalxy)
 {
     groups[group].decalxy = decalxy;
@@ -264,7 +440,9 @@ IdString Arch::wireToGlobal(int &row, int &col, const DatabasePOD *db, IdString
 {
     const std::string &wirename = wire.str(this);
     char buf[32];
-    if (wirename == "VCC" || wirename == "GND") {
+    if (wirename == "VCC" || wirename == "VSS") {
+        row = 0;
+        col = 0;
         return wire;
     }
     if (!isdigit(wirename[1]) || !isdigit(wirename[2]) || !isdigit(wirename[3])) {
@@ -505,10 +683,13 @@ static Loc getLoc(std::smatch match, int maxX, int maxY)
 
 void Arch::read_cst(std::istream &in)
 {
-    std::regex iobre = std::regex("IO_LOC +\"([^\"]+)\" +([^ ;]+) *;.*");
+    // If two locations are specified separated by commas (for differential I/O buffers),
+    // only the first location is actually recognized and used.
+    // And pin A will be Positive and pin B will be Negative in any case.
+    std::regex iobre = std::regex("IO_LOC +\"([^\"]+)\" +([^ ,;]+)(, *[^ ;]+)? *;.*");
     std::regex portre = std::regex("IO_PORT +\"([^\"]+)\" +([^;]+;).*");
     std::regex port_attrre = std::regex("([^ =;]+=[^ =;]+) *([^;]*;)");
-    std::regex iobelre = std::regex("IO([TRBL])([0-9]+)([A-Z])");
+    std::regex iobelre = std::regex("IO([TRBL])([0-9]+)\\[?([A-Z])\\]?");
     std::regex inslocre = std::regex("INS_LOC +\"([^\"]+)\" +R([0-9]+)C([0-9]+)\\[([0-9])\\]\\[([AB])\\] *;.*");
     std::smatch match, match_attr, match_pinloc;
     std::string line, pinline;
@@ -519,6 +700,7 @@ void Arch::read_cst(std::istream &in)
         insloc
     } cst_type;
 
+    settings.erase(id_cst);
     while (!in.eof()) {
         std::getline(in, line);
         cst_type = ioloc;
@@ -551,17 +733,19 @@ void Arch::read_cst(std::istream &in)
             if (belname != nullptr) {
                 std::string bel = IdString(belname->src_id).str(this);
                 it->second->setAttr(IdString(ID_BEL), bel);
-            } else if (std::regex_match(pinline, match_pinloc, iobelre)) {
-                // may be it's IOx#[AB] style?
-                Loc loc = getLoc(match_pinloc, getGridDimX(), getGridDimY());
-                BelId bel = getBelByLocation(loc);
-                if (bel == BelId()) {
-                    log_error("Pin %s not found\n", pinline.c_str());
-                }
-                std::string belname = getCtx()->nameOfBel(bel);
-                it->second->setAttr(IdString(ID_BEL), belname);
             } else {
-                log_error("Pin %s not found\n", pinname.c_str(this));
+                if (std::regex_match(pinline, match_pinloc, iobelre)) {
+                    // may be it's IOx#[AB] style?
+                    Loc loc = getLoc(match_pinloc, getGridDimX(), getGridDimY());
+                    BelId bel = getBelByLocation(loc);
+                    if (bel == BelId()) {
+                        log_error("Pin %s not found (TRBL style). \n", pinline.c_str());
+                    }
+                    std::string belname = getCtx()->nameOfBel(bel);
+                    it->second->setAttr(IdString(ID_BEL), belname);
+                } else {
+                    log_error("Pin %s not found (pin# style)\n", pinname.c_str(this));
+                }
             }
         } break;
         case insloc: { // INS_LOC
@@ -584,6 +768,7 @@ void Arch::read_cst(std::istream &in)
         }
         }
     }
+    settings[id_cst] = 1;
 }
 
 // Add all MUXes for the cell
@@ -606,13 +791,13 @@ void Arch::addMuxBels(const DatabasePOD *db, int row, int col)
 
     // 4 MUX2_LUT5, 2 MUX2_LUT6, 1 MUX2_LUT7, 1 MUX2_LUT8
     for (int j = 0; j < 8; ++j) {
-        z = j + mux_0_z;
+        z = j + BelZ::mux_0_z;
 
         int grow = row + 1;
         int gcol = col + 1;
 
         // no MUX2_LUT8 in the last column
-        if (j == 7 && col == getGridDimX() - 1) {
+        if (j == 7 && col == getGridDimX() - 2) {
             continue;
         }
 
@@ -689,6 +874,14 @@ Arch::Arch(ArchArgs args) : args(args)
         IdString::initialize_add(this, db->id_strs[i].get(), uint32_t(i) + db->num_constids);
     }
 
+    // Empty decal
+    addDecalGraphic(IdString(), GraphicElement());
+
+    if (args.gui) {
+        // decals
+        gfxCreateBelDecals(this);
+    }
+
     // setup package
     IdString package_name;
     IdString device_id;
@@ -705,7 +898,7 @@ Arch::Arch(ArchArgs args) : args(args)
         }
     }
     if (package_name == IdString()) {
-        log_error("Unsuported partnumber '%s'.\n", args.partnumber.c_str());
+        log_error("Unsupported partnumber '%s'.\n", args.partnumber.c_str());
     }
 
     // setup timing info
@@ -719,7 +912,7 @@ Arch::Arch(ArchArgs args) : args(args)
         }
     }
     if (speed == nullptr) {
-        log_error("Unsuported speed grade '%s'.\n", speed_id.c_str(this));
+        log_error("Unsupported speed grade '%s'.\n", speed_id.c_str(this));
     }
 
     const VariantPOD *variant = nullptr;
@@ -732,7 +925,7 @@ Arch::Arch(ArchArgs args) : args(args)
         }
     }
     if (variant == nullptr) {
-        log_error("Unsuported device grade '%s'.\n", device_id.c_str(this));
+        log_error("Unsupported device grade '%s'.\n", device_id.c_str(this));
     }
 
     package = nullptr;
@@ -750,7 +943,7 @@ Arch::Arch(ArchArgs args) : args(args)
     }
 
     if (package == nullptr) {
-        log_error("Unsuported package '%s'.\n", package_name.c_str(this));
+        log_error("Unsupported package '%s'.\n", package_name.c_str(this));
     }
 
     //
@@ -758,13 +951,28 @@ Arch::Arch(ArchArgs args) : args(args)
              package_name.c_str(this), speed_id.c_str(this));
 
     // setup db
+    // add global VCC and GND bels
+    addBel(id_GND, id_GND, Loc(0, 0, BelZ::gnd_0_z), true);
+    addWire(id_VSS, id_VSS, 0, 0);
+    addBelOutput(id_GND, id_G, id_VSS);
+    addBel(id_VCC, id_VCC, Loc(0, 0, BelZ::vcc_0_z), true);
+    addWire(id_VCC, id_VCC, 0, 0);
+    addBelOutput(id_VCC, id_V, id_VCC);
     char buf[32];
     // The reverse order of the enumeration simplifies the creation
     // of MUX2_LUT8s: they need the existence of the wire on the right.
     for (int i = db->rows * db->cols - 1; i >= 0; --i) {
+        IdString grpname;
         int row = i / db->cols;
         int col = i % db->cols;
         const TilePOD *tile = db->grid[i].get();
+        if (args.gui) {
+            // CRU decal
+            snprintf(buf, 32, "R%dC%d_CRU", row + 1, col + 1);
+            grpname = id(buf);
+            addGroup(grpname);
+            setGroupDecal(grpname, gfxGetCruGroupDecalXY(col, row));
+        }
         // setup wires
         const PairPOD *pips[2] = {tile->pips.get(), tile->clock_pips.get()};
         unsigned int num_pips[2] = {tile->num_pips, tile->num_clock_pips};
@@ -791,7 +999,54 @@ Arch::Arch(ArchArgs args) : args(args)
             IdString portname;
             int z = 0;
             bool dff = true;
+            bool oddrc = false;
             switch (static_cast<ConstIds>(bel->type_id)) {
+            case ID_GSR0:
+                snprintf(buf, 32, "R%dC%d_GSR0", row + 1, col + 1);
+                belname = id(buf);
+                addBel(belname, id_GSR, Loc(col, row, 0), false);
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_GSRI)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_GSRI, id(buf));
+                break;
+            case ID_OSC:
+                snprintf(buf, 32, "R%dC%d_OSC", row + 1, col + 1);
+                belname = id(buf);
+                addBel(belname, id_OSC, Loc(col, row, BelZ::osc_z), false);
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_OSCOUT)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelOutput(belname, id_OSCOUT, id(buf));
+                break;
+            case ID_OSCH:
+                snprintf(buf, 32, "R%dC%d_OSCH", row + 1, col + 1);
+                belname = id(buf);
+                addBel(belname, id_OSCH, Loc(col, row, BelZ::osc_z), false);
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_OSCOUT)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelOutput(belname, id_OSCOUT, id(buf));
+                break;
+            case ID_OSCF:
+                snprintf(buf, 32, "R%dC%d_OSCF", row + 1, col + 1);
+                belname = id(buf);
+                addBel(belname, id_OSCF, Loc(col, row, BelZ::osc_z), false);
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_OSCOUT)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelOutput(belname, id_OSCOUT, id(buf));
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_OSCEN)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_OSCEN, id(buf));
+                break;
+            case ID_OSCZ:
+                snprintf(buf, 32, "R%dC%d_OSCZ", row + 1, col + 1);
+                belname = id(buf);
+                addBel(belname, id_OSCZ, Loc(col, row, BelZ::osc_z), false);
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_OSCOUT)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelOutput(belname, id_OSCOUT, id(buf));
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_OSCEN)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_OSCEN, id(buf));
+                break;
             // fall through the ++
             case ID_LUT7:
                 z++;
@@ -837,6 +1092,14 @@ Arch::Arch(ArchArgs args) : args(args)
                 if (z == 0) {
                     addMuxBels(db, row, col);
                 }
+                if (z % 2 == 0) {
+                    snprintf(buf, 32, "R%dC%d_LUT_GRP%d", row + 1, col + 1, z);
+                    grpname = id(buf);
+                    if (args.gui) {
+                        addGroup(grpname);
+                        setGroupDecal(grpname, gfxGetLutGroupDecalXY(col, row, z >> 1));
+                    }
+                }
                 break;
             case ID_IOBJ:
                 z++; /* fall-through*/
@@ -856,7 +1119,7 @@ Arch::Arch(ArchArgs args) : args(args)
                 z++; /* fall-through*/
             case ID_IOBB:
                 z++; /* fall-through*/
-            case ID_IOBA:
+            case ID_IOBA: {
                 snprintf(buf, 32, "R%dC%d_IOB%c", row + 1, col + 1, 'A' + z);
                 belname = id(buf);
                 addBel(belname, id_IOB, Loc(col, row, z), false);
@@ -869,8 +1132,121 @@ Arch::Arch(ArchArgs args) : args(args)
                 portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_OE)->src_id);
                 snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
                 addBelInput(belname, id_OEN, id(buf));
+                // GW1NR-9 quirk
+                const PairPOD *xxx_port = pairLookup(bel->ports.get(), bel->num_ports, ID_XXX_VSS0);
+                if (xxx_port != nullptr) {
+                    gw1n9_quirk = true;
+                    portname = IdString(xxx_port->src_id);
+                    snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                    addBelInput(belname, id_XXX_VSS0, id(buf));
+                }
+                xxx_port = pairLookup(bel->ports.get(), bel->num_ports, ID_XXX_VSS1);
+                if (xxx_port != nullptr) {
+                    gw1n9_quirk = true;
+                    portname = IdString(xxx_port->src_id);
+                    snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                    addBelInput(belname, id_XXX_VSS1, id(buf));
+                }
+            } break;
+                // Simplified IO
+            case ID_IOBJS:
+                z++; /* fall-through*/
+            case ID_IOBIS:
+                z++; /* fall-through*/
+            case ID_IOBHS:
+                z++; /* fall-through*/
+            case ID_IOBGS:
+                z++; /* fall-through*/
+            case ID_IOBFS:
+                z++; /* fall-through*/
+            case ID_IOBES:
+                z++; /* fall-through*/
+            case ID_IOBDS:
+                z++; /* fall-through*/
+            case ID_IOBCS:
+                z++; /* fall-through*/
+            case ID_IOBBS:
+                z++; /* fall-through*/
+            case ID_IOBAS:
+                snprintf(buf, 32, "R%dC%d_IOB%c", row + 1, col + 1, 'A' + z);
+                belname = id(buf);
+                addBel(belname, id_IOBS, Loc(col, row, z), false);
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_O)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelOutput(belname, id_O, id(buf));
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_I)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_I, id(buf));
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_OE)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_OEN, id(buf));
                 break;
 
+                // IO logic
+            case ID_ODDRCB:
+                z++; /* fall-through*/
+            case ID_ODDRCA:
+                oddrc = true;
+                z++; /* fall-through*/
+            case ID_ODDRB:
+                z++; /* fall-through*/
+            case ID_ODDRA: {
+                snprintf(buf, 32, "R%dC%d_ODDR%s%c", row + 1, col + 1, oddrc ? "C" : "", 'A' + z - (oddrc ? 2 : 0));
+                belname = id(buf);
+                addBel(belname, id_ODDR, Loc(col, row, BelZ::iologic_0_z + z), false);
+
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_D0)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_D0, id(buf));
+
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_D1)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_D1, id(buf));
+
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_TX)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_TX, id(buf));
+
+                portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_CLK)->src_id);
+                snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                addBelInput(belname, id_CLK, id(buf));
+
+                const PairPOD *xxx_port = pairLookup(bel->ports.get(), bel->num_ports, ID_XXX_VSS);
+                if (xxx_port != nullptr) {
+                    ddr_has_extra_inputs = true;
+                    portname = IdString(xxx_port->src_id);
+                    snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                    addBelInput(belname, id_XXX_VSS, id(buf));
+                }
+                xxx_port = pairLookup(bel->ports.get(), bel->num_ports, ID_XXX_VCC);
+                if (xxx_port != nullptr) {
+                    ddr_has_extra_inputs = true;
+                    portname = IdString(xxx_port->src_id);
+                    snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                    addBelInput(belname, id_XXX_VCC, id(buf));
+                }
+
+                if (oddrc) {
+                    portname = IdString(pairLookup(bel->ports.get(), bel->num_ports, ID_CE)->src_id);
+                    snprintf(buf, 32, "R%dC%d_%s", row + 1, col + 1, portname.c_str(this));
+                    addBelInput(belname, id_CE, id(buf));
+                }
+
+                // dummy wires
+                snprintf(buf, 32, "ODDR%s%c_Q0", oddrc ? "C" : "", 'A' + z - (oddrc ? 2 : 0));
+                IdString id_q0 = id(buf);
+                IdString q0_name = wireToGlobal(row, col, db, id_q0);
+                if (wires.count(q0_name) == 0)
+                    addWire(q0_name, id_q0, row, col);
+                addBelOutput(belname, id_Q0, q0_name);
+
+                snprintf(buf, 32, "ODDR%s%c_Q1", oddrc ? "C" : "", 'A' + z - (oddrc ? 2 : 0));
+                IdString id_q1 = id(buf);
+                IdString q1_name = wireToGlobal(row, col, db, id_q1);
+                if (wires.count(q1_name) == 0)
+                    addWire(q1_name, id_q1, row, col);
+                addBelOutput(belname, id_Q1, q1_name);
+            } break;
             default:
                 break;
             }
@@ -900,7 +1276,6 @@ Arch::Arch(ArchArgs args) : args(args)
                 DelayQuad delay = getWireTypeDelay(destid);
                 // local alias
                 auto local_alias = pairLookup(tile->aliases.get(), tile->num_aliases, srcid.index);
-                // std::cout << "srcid " << srcid.str(this) << std::endl;
                 if (local_alias != nullptr) {
                     srcid = IdString(local_alias->src_id);
                     gsrcname = wireToGlobal(srcrow, srccol, db, srcid);
@@ -917,12 +1292,14 @@ Arch::Arch(ArchArgs args) : args(args)
                     srcrow = alias_src->src_row;
                     srcid = IdString(alias_src->src_id);
                     gsrcname = wireToGlobal(srcrow, srccol, db, srcid);
-                    // std::cout << buf << std::endl;
                 }
                 addPip(pipname, destid, gsrcname, gdestname, delay, Loc(col, row, j));
             }
         }
     }
+    if (args.gui) {
+        setDefaultDecals();
+    }
 
     // Permissible combinations of modes in a single slice
     dff_comp_mode[id_DFF] = id_DFF;
@@ -982,6 +1359,17 @@ BelId Arch::getBelByLocation(Loc loc) const
 
 const std::vector<BelId> &Arch::getBelsByTile(int x, int y) const { return bels_by_tile.at(x).at(y); }
 
+bool Arch::haveBelType(int x, int y, IdString bel_type)
+{
+    for (auto bel : getBelsByTile(x, y)) {
+        BelInfo bi = bel_info(bel);
+        if (bi.type == bel_type) {
+            return true;
+        }
+    }
+    return false;
+}
+
 bool Arch::getBelGlobalBuf(BelId bel) const { return bels.at(bel).gb; }
 
 void Arch::bindBel(BelId bel, CellInfo *cell, PlaceStrength strength)
@@ -1171,11 +1559,12 @@ delay_t Arch::estimateDelay(WireId src, WireId dst) const
     return (dx + dy) * args.delayScale + args.delayOffset;
 }
 
-delay_t Arch::predictDelay(const NetInfo *net_info, const PortRef &sink) const
+delay_t Arch::predictDelay(BelId src_bel, IdString src_pin, BelId dst_bel, IdString dst_pin) const
 {
-    const auto &driver = net_info->driver;
-    auto driver_loc = getBelLocation(driver.cell->bel);
-    auto sink_loc = getBelLocation(sink.cell->bel);
+    NPNR_UNUSED(src_pin);
+    NPNR_UNUSED(dst_pin);
+    auto driver_loc = getBelLocation(src_bel);
+    auto sink_loc = getBelLocation(dst_bel);
 
     int dx = abs(sink_loc.x - driver_loc.x);
     int dy = abs(sink_loc.y - driver_loc.y);
@@ -1212,13 +1601,13 @@ ArcBounds Arch::getRouteBoundingBox(WireId src, WireId dst) const
 
 bool Arch::place()
 {
-    std::string placer = str_or_default(settings, id("placer"), defaultPlacer);
+    std::string placer = str_or_default(settings, id_placer, defaultPlacer);
     bool retVal;
     if (placer == "heap") {
         bool have_iobuf_or_constr = false;
         for (auto &cell : cells) {
             CellInfo *ci = cell.second.get();
-            if (ci->type == id("IOB") || ci->bel != BelId() || ci->attrs.count(id("BEL"))) {
+            if (ci->type == id_IOB || ci->bel != BelId() || ci->attrs.count(id_BEL)) {
                 have_iobuf_or_constr = true;
                 break;
             }
@@ -1229,15 +1618,15 @@ bool Arch::place()
             retVal = placer1(getCtx(), Placer1Cfg(getCtx()));
         } else {
             PlacerHeapCfg cfg(getCtx());
-            cfg.ioBufTypes.insert(id("IOB"));
+            cfg.ioBufTypes.insert(id_IOB);
             cfg.beta = 0.5;
             retVal = placer_heap(getCtx(), cfg);
         }
-        getCtx()->settings[getCtx()->id("place")] = 1;
+        getCtx()->settings[id_place] = 1;
         archInfoToAttributes();
     } else if (placer == "sa") {
         retVal = placer1(getCtx(), Placer1Cfg(getCtx()));
-        getCtx()->settings[getCtx()->id("place")] = 1;
+        getCtx()->settings[id_place] = 1;
         archInfoToAttributes();
         return retVal;
     } else {
@@ -1254,7 +1643,7 @@ bool Arch::place()
 
 bool Arch::route()
 {
-    std::string router = str_or_default(settings, id("router"), defaultRouter);
+    std::string router = str_or_default(settings, id_router, defaultRouter);
     bool result;
     if (router == "router1") {
         result = router1(getCtx(), Router1Cfg(getCtx()));
@@ -1264,11 +1653,21 @@ bool Arch::route()
     } else {
         log_error("Gowin architecture does not support router '%s'\n", router.c_str());
     }
-    getCtx()->settings[getCtx()->id("route")] = 1;
+    getCtx()->settings[id_route] = 1;
     archInfoToAttributes();
     return result;
 }
 
+// ---------------------------------------------------------------
+std::vector<GraphicElement> Arch::getDecalGraphics(DecalId decal) const
+{
+    if (!decal_graphics.count(decal)) {
+        // XXX
+        return std::vector<GraphicElement>();
+    }
+    return decal_graphics.at(decal);
+}
+
 // ---------------------------------------------------------------
 
 bool Arch::getCellDelay(const CellInfo *cell, IdString fromPort, IdString toPort, DelayQuad &delay) const
@@ -1347,9 +1746,9 @@ void Arch::assignArchInfo()
             ci->is_slice = true;
             ci->ff_used = ci->params.at(id_FF_USED).as_bool();
             ci->ff_type = id(ci->params.at(id_FF_TYPE).as_string());
-            ci->slice_clk = get_net_or_empty(ci, id("CLK"));
-            ci->slice_ce = get_net_or_empty(ci, id("CE"));
-            ci->slice_lsr = get_net_or_empty(ci, id("LSR"));
+            ci->slice_clk = ci->getPort(id_CLK);
+            ci->slice_ce = ci->getPort(id_CE);
+            ci->slice_lsr = ci->getPort(id_LSR);
 
             // add timing paths
             addCellTimingClock(cname, id_CLK);
