@@ -68,7 +68,7 @@ static void pack_alus(Context *ctx)
                 continue;
             }
 
-            if (!is_alu(ctx, cin_ci) || cin->users.size() > 1) {
+            if (!is_alu(ctx, cin_ci) || cin->users.entries() > 1) {
                 if (ctx->verbose) {
                     log_info("ALU head found %s. CIN net is %s\n", ctx->nameOf(ci), ctx->nameOf(cin));
                 }
@@ -89,11 +89,14 @@ static void pack_alus(Context *ctx)
         }
         std::unique_ptr<CellInfo> packed_head = create_generic_cell(ctx, id_SLICE, ci->name.str(ctx) + "_HEAD_ALULC");
 
+        // Head is always SLICE0
+        packed_head->constr_z = 0;
+        packed_head->constr_abs_z = true;
         if (ctx->verbose) {
             log_info("packed ALU head into %s. CIN net is %s\n", ctx->nameOf(packed_head.get()),
                      ctx->nameOf(cin_netId));
         }
-        connect_port(ctx, ctx->nets[ctx->id("$PACKER_VCC_NET")].get(), packed_head.get(), id_C);
+        packed_head->connectPort(id_C, ctx->nets[ctx->id("$PACKER_VCC_NET")].get());
         if (cin_netId == ctx->id("$PACKER_GND_NET")) {
             // CIN = 0
             packed_head->params[id_ALU_MODE] = std::string("C2L");
@@ -103,15 +106,15 @@ static void pack_alus(Context *ctx)
                 packed_head->params[id_ALU_MODE] = std::string("ONE2C");
             } else {
                 // CIN from logic
-                connect_port(ctx, ctx->nets[cin_netId].get(), packed_head.get(), id_B);
-                connect_port(ctx, ctx->nets[cin_netId].get(), packed_head.get(), id_D);
+                packed_head->connectPort(id_B, ctx->nets[cin_netId].get());
+                packed_head->connectPort(id_D, ctx->nets[cin_netId].get());
                 packed_head->params[id_ALU_MODE] = std::string("0"); // ADD
             }
         }
 
         int alu_idx = 1;
         do { // go through the ALU chain
-            auto alu_bel = ci->attrs.find(ctx->id("BEL"));
+            auto alu_bel = ci->attrs.find(id_BEL);
             if (alu_bel != ci->attrs.end()) {
                 log_error("ALU %s placement restrictions are not supported.\n", ctx->nameOf(ci));
                 return;
@@ -120,9 +123,9 @@ static void pack_alus(Context *ctx)
             packed_cells.insert(ci->name);
 
             // CIN/COUT are hardwired, delete
-            disconnect_port(ctx, ci, id_CIN);
+            ci->disconnectPort(id_CIN);
             NetInfo *cout = ci->ports.at(id_COUT).net;
-            disconnect_port(ctx, ci, id_COUT);
+            ci->disconnectPort(id_COUT);
 
             std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, id_SLICE, ci->name.str(ctx) + "_ALULC");
             if (ctx->verbose) {
@@ -132,9 +135,9 @@ static void pack_alus(Context *ctx)
             int mode = int_or_default(ci->params, id_ALU_MODE);
             packed->params[id_ALU_MODE] = mode;
             if (mode == 9) { // MULT
-                connect_port(ctx, ctx->nets[ctx->id("$PACKER_GND_NET")].get(), packed.get(), id_C);
+                packed->connectPort(id_C, ctx->nets[ctx->id("$PACKER_GND_NET")].get());
             } else {
-                connect_port(ctx, ctx->nets[ctx->id("$PACKER_VCC_NET")].get(), packed.get(), id_C);
+                packed->connectPort(id_C, ctx->nets[ctx->id("$PACKER_VCC_NET")].get());
             }
 
             // add to cluster
@@ -146,37 +149,37 @@ static void pack_alus(Context *ctx)
             ++alu_idx;
 
             // connect all remainig ports
-            replace_port(ci, id_SUM, packed.get(), id_F);
+            ci->movePortTo(id_SUM, packed.get(), id_F);
             switch (mode) {
             case 0: // ADD
-                replace_port(ci, id_I0, packed.get(), id_B);
-                replace_port(ci, id_I1, packed.get(), id_D);
+                ci->movePortTo(id_I0, packed.get(), id_B);
+                ci->movePortTo(id_I1, packed.get(), id_D);
                 break;
             case 1: // SUB
-                replace_port(ci, id_I0, packed.get(), id_A);
-                replace_port(ci, id_I1, packed.get(), id_D);
+                ci->movePortTo(id_I0, packed.get(), id_A);
+                ci->movePortTo(id_I1, packed.get(), id_D);
                 break;
             case 5: // LE
-                replace_port(ci, id_I0, packed.get(), id_A);
-                replace_port(ci, id_I1, packed.get(), id_B);
+                ci->movePortTo(id_I0, packed.get(), id_A);
+                ci->movePortTo(id_I1, packed.get(), id_B);
                 break;
             case 9: // MULT
-                replace_port(ci, id_I0, packed.get(), id_A);
-                replace_port(ci, id_I1, packed.get(), id_B);
-                disconnect_port(ctx, packed.get(), id_D);
-                connect_port(ctx, ctx->nets[ctx->id("$PACKER_VCC_NET")].get(), packed.get(), id_D);
+                ci->movePortTo(id_I0, packed.get(), id_A);
+                ci->movePortTo(id_I1, packed.get(), id_B);
+                packed->disconnectPort(id_D);
+                packed->connectPort(id_D, ctx->nets[ctx->id("$PACKER_VCC_NET")].get());
                 break;
             default:
-                replace_port(ci, id_I0, packed.get(), id_A);
-                replace_port(ci, id_I1, packed.get(), id_B);
-                replace_port(ci, id_I3, packed.get(), id_D);
+                ci->movePortTo(id_I0, packed.get(), id_A);
+                ci->movePortTo(id_I1, packed.get(), id_B);
+                ci->movePortTo(id_I3, packed.get(), id_D);
             }
 
             new_cells.push_back(std::move(packed));
 
-            if (cout != nullptr && cout->users.size() > 0) {
+            if (cout != nullptr && cout->users.entries() > 0) {
                 // if COUT used by logic
-                if ((cout->users.size() > 1) || (!is_alu(ctx, cout->users.at(0).cell))) {
+                if ((cout->users.entries() > 1) || (!is_alu(ctx, (*cout->users.begin()).cell))) {
                     if (ctx->verbose) {
                         log_info("COUT is used by logic\n");
                     }
@@ -188,7 +191,7 @@ static void pack_alus(Context *ctx)
                                  ctx->nameOf(cout));
                     }
                     packed_tail->params[id_ALU_MODE] = std::string("C2L");
-                    connect_port(ctx, cout, packed_tail.get(), id_F);
+                    packed_tail->connectPort(id_F, cout);
                     // add to cluster
                     packed_tail->cluster = packed_head->name;
                     packed_tail->constr_z = alu_idx % 6;
@@ -201,7 +204,7 @@ static void pack_alus(Context *ctx)
                     break;
                 }
                 // next ALU
-                ci = cout->users.at(0).cell;
+                ci = (*cout->users.begin()).cell;
                 // if ALU is too big
                 if (alu_idx == (ctx->gridDimX - 2) * 6 - 1) {
                     log_error("ALU %s is the %dth in the chain. Such long chains are not supported.\n", ctx->nameOf(ci),
@@ -240,7 +243,7 @@ static void pack_mux2_lut5(Context *ctx, CellInfo *ci, pool<IdString> &packed_ce
                            std::vector<std::unique_ptr<CellInfo>> &new_cells)
 {
 
-    if (bool_or_default(ci->attrs, ctx->id("SINGLE_INPUT_MUX"))) {
+    if (bool_or_default(ci->attrs, id_SINGLE_INPUT_MUX)) {
         // find the muxed LUT
         NetInfo *i1 = ci->ports.at(id_I1).net;
 
@@ -254,26 +257,26 @@ static void pack_mux2_lut5(Context *ctx, CellInfo *ci, pool<IdString> &packed_ce
         }
 
         // XXX enable the placement constraints
-        auto mux_bel = ci->attrs.find(ctx->id("BEL"));
-        auto lut1_bel = lut1->attrs.find(ctx->id("BEL"));
+        auto mux_bel = ci->attrs.find(id_BEL);
+        auto lut1_bel = lut1->attrs.find(id_BEL);
         if (lut1_bel != lut1->attrs.end() || mux_bel != ci->attrs.end()) {
             log_error("MUX2_LUT5 '%s' placement restrictions are not supported yet\n", ctx->nameOf(ci));
             return;
         }
 
-        std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, ctx->id("GW_MUX2_LUT5"), ci->name.str(ctx) + "_LC");
+        std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, id_GW_MUX2_LUT5, ci->name.str(ctx) + "_LC");
         if (ctx->verbose) {
             log_info("packed cell %s into %s\n", ctx->nameOf(ci), ctx->nameOf(packed.get()));
         }
         // mux is the cluster root
         packed->cluster = packed->name;
         lut1->cluster = packed->name;
-        lut1->constr_z = -ctx->mux_0_z + 1;
+        lut1->constr_z = -BelZ::mux_0_z + 1;
         packed->constr_children.clear();
 
         // reconnect MUX ports
-        replace_port(ci, id_O, packed.get(), id_OF);
-        replace_port(ci, id_I1, packed.get(), id_I1);
+        ci->movePortTo(id_O, packed.get(), id_OF);
+        ci->movePortTo(id_I1, packed.get(), id_I1);
 
         // remove cells
         packed_cells.insert(ci->name);
@@ -296,31 +299,31 @@ static void pack_mux2_lut5(Context *ctx, CellInfo *ci, pool<IdString> &packed_ce
         }
 
         // XXX enable the placement constraints
-        auto mux_bel = ci->attrs.find(ctx->id("BEL"));
-        auto lut0_bel = lut0->attrs.find(ctx->id("BEL"));
-        auto lut1_bel = lut1->attrs.find(ctx->id("BEL"));
+        auto mux_bel = ci->attrs.find(id_BEL);
+        auto lut0_bel = lut0->attrs.find(id_BEL);
+        auto lut1_bel = lut1->attrs.find(id_BEL);
         if (lut0_bel != lut0->attrs.end() || lut1_bel != lut1->attrs.end() || mux_bel != ci->attrs.end()) {
             log_error("MUX2_LUT5 '%s' placement restrictions are not supported yet\n", ctx->nameOf(ci));
             return;
         }
 
-        std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, ctx->id("GW_MUX2_LUT5"), ci->name.str(ctx) + "_LC");
+        std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, id_GW_MUX2_LUT5, ci->name.str(ctx) + "_LC");
         if (ctx->verbose) {
             log_info("packed cell %s into %s\n", ctx->nameOf(ci), ctx->nameOf(packed.get()));
         }
         // mux is the cluster root
         packed->cluster = packed->name;
         lut0->cluster = packed->name;
-        lut0->constr_z = -ctx->mux_0_z;
+        lut0->constr_z = -BelZ::mux_0_z;
         lut1->cluster = packed->name;
-        lut1->constr_z = -ctx->mux_0_z + 1;
+        lut1->constr_z = -BelZ::mux_0_z + 1;
         packed->constr_children.clear();
 
         // reconnect MUX ports
-        replace_port(ci, id_O, packed.get(), id_OF);
-        replace_port(ci, id_S0, packed.get(), id_SEL);
-        replace_port(ci, id_I0, packed.get(), id_I0);
-        replace_port(ci, id_I1, packed.get(), id_I1);
+        ci->movePortTo(id_O, packed.get(), id_OF);
+        ci->movePortTo(id_S0, packed.get(), id_SEL);
+        ci->movePortTo(id_I0, packed.get(), id_I0);
+        ci->movePortTo(id_I1, packed.get(), id_I1);
 
         // remove cells
         packed_cells.insert(ci->name);
@@ -351,9 +354,9 @@ static void pack_mux2_lut(Context *ctx, CellInfo *ci, bool (*pred)(const BaseCtx
     }
 
     // XXX enable the placement constraints
-    auto mux_bel = ci->attrs.find(ctx->id("BEL"));
-    auto mux0_bel = mux0->attrs.find(ctx->id("BEL"));
-    auto mux1_bel = mux1->attrs.find(ctx->id("BEL"));
+    auto mux_bel = ci->attrs.find(id_BEL);
+    auto mux0_bel = mux0->attrs.find(id_BEL);
+    auto mux1_bel = mux1->attrs.find(id_BEL);
     if (mux0_bel != mux0->attrs.end() || mux1_bel != mux1->attrs.end() || mux_bel != ci->attrs.end()) {
         log_error("MUX2_LUT%c '%s' placement restrictions are not supported yet\n", type_suffix, ctx->nameOf(ci));
         return;
@@ -391,10 +394,10 @@ static void pack_mux2_lut(Context *ctx, CellInfo *ci, bool (*pred)(const BaseCtx
     packed->constr_children.push_back(mux1);
 
     // reconnect MUX ports
-    replace_port(ci, id_O, packed.get(), id_OF);
-    replace_port(ci, id_S0, packed.get(), id_SEL);
-    replace_port(ci, id_I0, packed.get(), id_I0);
-    replace_port(ci, id_I1, packed.get(), id_I1);
+    ci->movePortTo(id_O, packed.get(), id_OF);
+    ci->movePortTo(id_S0, packed.get(), id_SEL);
+    ci->movePortTo(id_I0, packed.get(), id_I0);
+    ci->movePortTo(id_I1, packed.get(), id_I1);
 
     // remove cells
     packed_cells.insert(ci->name);
@@ -509,7 +512,7 @@ static void pack_lut_lutffs(Context *ctx)
         if (ctx->verbose)
             log_info("cell '%s' is of type '%s'\n", ctx->nameOf(ci), ci->type.c_str(ctx));
         if (is_lut(ctx, ci)) {
-            std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, ctx->id("SLICE"), ci->name.str(ctx) + "_LC");
+            std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, id_SLICE, ci->name.str(ctx) + "_LC");
             for (auto &attr : ci->attrs)
                 packed->attrs[attr.first] = attr.second;
             packed_cells.insert(ci->name);
@@ -517,14 +520,14 @@ static void pack_lut_lutffs(Context *ctx)
                 log_info("packed cell %s into %s\n", ctx->nameOf(ci), ctx->nameOf(packed.get()));
             // See if we can pack into a DFF
             // TODO: LUT cascade
-            NetInfo *o = ci->ports.at(ctx->id("F")).net;
-            CellInfo *dff = net_only_drives(ctx, o, is_ff, ctx->id("D"), true);
-            auto lut_bel = ci->attrs.find(ctx->id("BEL"));
+            NetInfo *o = ci->ports.at(id_F).net;
+            CellInfo *dff = net_only_drives(ctx, o, is_ff, id_D, true);
+            auto lut_bel = ci->attrs.find(id_BEL);
             bool packed_dff = false;
             if (dff) {
                 if (ctx->verbose)
                     log_info("found attached dff %s\n", ctx->nameOf(dff));
-                auto dff_bel = dff->attrs.find(ctx->id("BEL"));
+                auto dff_bel = dff->attrs.find(id_BEL);
                 if (lut_bel != ci->attrs.end() && dff_bel != dff->attrs.end() && lut_bel->second != dff_bel->second) {
                     // Locations don't match, can't pack
                 } else {
@@ -532,7 +535,7 @@ static void pack_lut_lutffs(Context *ctx)
                     dff_to_lc(ctx, dff, packed.get(), false);
                     ctx->nets.erase(o->name);
                     if (dff_bel != dff->attrs.end())
-                        packed->attrs[ctx->id("BEL")] = dff_bel->second;
+                        packed->attrs[id_BEL] = dff_bel->second;
                     packed_cells.insert(dff->name);
                     if (ctx->verbose)
                         log_info("packed cell %s into %s\n", ctx->nameOf(dff), ctx->nameOf(packed.get()));
@@ -564,7 +567,7 @@ static void pack_nonlut_ffs(Context *ctx)
     for (auto &cell : ctx->cells) {
         CellInfo *ci = cell.second.get();
         if (is_ff(ctx, ci)) {
-            std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, ctx->id("SLICE"), ci->name.str(ctx) + "_DFFLC");
+            std::unique_ptr<CellInfo> packed = create_generic_cell(ctx, id_SLICE, ci->name.str(ctx) + "_DFFLC");
             for (auto &attr : ci->attrs)
                 packed->attrs[attr.first] = attr.second;
             if (ctx->verbose)
@@ -593,9 +596,10 @@ static void set_net_constant(const Context *ctx, NetInfo *orig, NetInfo *constne
                 log_info("%s user %s\n", ctx->nameOf(orig), ctx->nameOf(uc));
             if ((is_lut(ctx, uc) || is_lc(ctx, uc)) && (user.port.str(ctx).at(0) == 'I') && !constval) {
                 uc->ports[user.port].net = nullptr;
+                uc->ports[user.port].user_idx = {};
             } else {
                 uc->ports[user.port].net = constnet;
-                constnet->users.push_back(user);
+                uc->ports[user.port].user_idx = constnet->users.add(user);
             }
         }
     }
@@ -607,36 +611,31 @@ static void pack_constants(Context *ctx)
 {
     log_info("Packing constants..\n");
 
-    std::unique_ptr<CellInfo> gnd_cell = create_generic_cell(ctx, ctx->id("SLICE"), "$PACKER_GND");
-    gnd_cell->params[ctx->id("INIT")] = Property(0, 1 << 4);
-    std::unique_ptr<NetInfo> gnd_net = std::unique_ptr<NetInfo>(new NetInfo);
-    gnd_net->name = ctx->id("$PACKER_GND_NET");
+    std::unique_ptr<CellInfo> gnd_cell = create_generic_cell(ctx, id_GND, "$PACKER_GND");
+    auto gnd_net = std::make_unique<NetInfo>(ctx->id("$PACKER_GND_NET"));
     gnd_net->driver.cell = gnd_cell.get();
-    gnd_net->driver.port = ctx->id("F");
-    gnd_cell->ports.at(ctx->id("F")).net = gnd_net.get();
+    gnd_net->driver.port = id_G;
+    gnd_cell->ports.at(id_G).net = gnd_net.get();
 
-    std::unique_ptr<CellInfo> vcc_cell = create_generic_cell(ctx, ctx->id("SLICE"), "$PACKER_VCC");
-    // Fill with 1s
-    vcc_cell->params[ctx->id("INIT")] = Property(Property::S1).extract(0, (1 << 4), Property::S1);
-    std::unique_ptr<NetInfo> vcc_net = std::unique_ptr<NetInfo>(new NetInfo);
-    vcc_net->name = ctx->id("$PACKER_VCC_NET");
+    std::unique_ptr<CellInfo> vcc_cell = create_generic_cell(ctx, id_VCC, "$PACKER_VCC");
+    auto vcc_net = std::make_unique<NetInfo>(ctx->id("$PACKER_VCC_NET"));
     vcc_net->driver.cell = vcc_cell.get();
-    vcc_net->driver.port = ctx->id("F");
-    vcc_cell->ports.at(ctx->id("F")).net = vcc_net.get();
+    vcc_net->driver.port = id_V;
+    vcc_cell->ports.at(id_V).net = vcc_net.get();
 
     std::vector<IdString> dead_nets;
 
-    bool gnd_used = false;
+    bool gnd_used = true; // XXX May be needed for simplified IO
 
     for (auto &net : ctx->nets) {
         NetInfo *ni = net.second.get();
-        if (ni->driver.cell != nullptr && ni->driver.cell->type == ctx->id("GND")) {
+        if (ni->driver.cell != nullptr && ni->driver.cell->type == id_GND) {
             IdString drv_cell = ni->driver.cell->name;
             set_net_constant(ctx, ni, gnd_net.get(), false);
             gnd_used = true;
             dead_nets.push_back(net.first);
             ctx->cells.erase(drv_cell);
-        } else if (ni->driver.cell != nullptr && ni->driver.cell->type == ctx->id("VCC")) {
+        } else if (ni->driver.cell != nullptr && ni->driver.cell->type == id_VCC) {
             IdString drv_cell = ni->driver.cell->name;
             set_net_constant(ctx, ni, vcc_net.get(), true);
             dead_nets.push_back(net.first);
@@ -658,6 +657,41 @@ static void pack_constants(Context *ctx)
     }
 }
 
+// Pack global set-reset
+static void pack_gsr(Context *ctx)
+{
+    log_info("Packing GSR..\n");
+
+    bool user_gsr = false;
+    for (auto &cell : ctx->cells) {
+        CellInfo *ci = cell.second.get();
+        if (ctx->verbose)
+            log_info("cell '%s' is of type '%s'\n", ctx->nameOf(ci), ci->type.c_str(ctx));
+        if (ci->type == id_GSR) {
+            user_gsr = true;
+            break;
+        }
+    }
+    if (!user_gsr) {
+        // XXX
+        bool have_gsr_bel = false;
+        for (auto bi : ctx->bels) {
+            if (bi.second.type == id_GSR) {
+                have_gsr_bel = true;
+                break;
+            }
+        }
+        if (have_gsr_bel) {
+            // make default GSR
+            std::unique_ptr<CellInfo> gsr_cell = create_generic_cell(ctx, id_GSR, "GSR");
+            gsr_cell->connectPort(id_GSRI, ctx->nets[ctx->id("$PACKER_VCC_NET")].get());
+            ctx->cells[gsr_cell->name] = std::move(gsr_cell);
+        } else {
+            log_info("No GSR in the chip base\n");
+        }
+    }
+}
+
 static bool is_nextpnr_iob(const Context *ctx, CellInfo *cell)
 {
     return cell->type == ctx->id("$nextpnr_ibuf") || cell->type == ctx->id("$nextpnr_obuf") ||
@@ -677,6 +711,193 @@ static bool is_gowin_iob(const Context *ctx, const CellInfo *cell)
     }
 }
 
+static bool is_gowin_diff_iob(const Context *ctx, const CellInfo *cell)
+{
+    switch (cell->type.index) {
+    case ID_TLVDS_OBUF:
+        return true;
+    default:
+        return false;
+    }
+}
+
+static bool is_gowin_iologic(const Context *ctx, const CellInfo *cell)
+{
+    switch (cell->type.index) {
+    case ID_ODDR: /* fall-through*/
+    case ID_ODDRC:
+        return true;
+    default:
+        return false;
+    }
+}
+
+static bool is_iob(const Context *ctx, const CellInfo *cell) { return (cell->type.index == ID_IOB); }
+
+// Pack IO logic
+static void pack_iologic(Context *ctx)
+{
+    pool<IdString> packed_cells;
+    pool<IdString> delete_nets;
+
+    std::vector<std::unique_ptr<CellInfo>> new_cells;
+    log_info("Packing IO logic..\n");
+
+    for (auto &cell : ctx->cells) {
+        CellInfo *ci = cell.second.get();
+        if (ctx->verbose)
+            log_info("cell '%s' is of type '%s'\n", ctx->nameOf(ci), ci->type.c_str(ctx));
+        if (is_gowin_iologic(ctx, ci)) {
+            CellInfo *q0_dst = nullptr;
+            CellInfo *q1_dst = nullptr;
+            switch (ci->type.index) {
+            case ID_ODDRC: /* fall-through*/
+            case ID_ODDR: {
+                q0_dst = net_only_drives(ctx, ci->ports.at(id_Q0).net, is_iob, id_I);
+                NPNR_ASSERT(q0_dst != nullptr);
+
+                auto iob_bel = q0_dst->attrs.find(id_BEL);
+                if (q0_dst->attrs.count(id_DIFF_TYPE)) {
+                    ci->attrs[id_OBUF_TYPE] = std::string("DBUF");
+                } else {
+                    ci->attrs[id_OBUF_TYPE] = std::string("SBUF");
+                }
+                if (iob_bel != q0_dst->attrs.end()) {
+                    // already know there to place, no need of any cluster stuff
+                    Loc loc = ctx->getBelLocation(ctx->getBelByNameStr(iob_bel->second.as_string()));
+                    loc.z += BelZ::iologic_0_z;
+                    ci->attrs[id_BEL] = ctx->getBelName(ctx->getBelByLocation(loc)).str(ctx);
+                } else {
+                    // make cluster from ODDR and OBUF
+                    ci->cluster = ci->name;
+                    ci->constr_x = 0;
+                    ci->constr_y = 0;
+                    ci->constr_z = 0;
+                    ci->constr_abs_z = false;
+                    ci->constr_children.push_back(q0_dst);
+                    q0_dst->cluster = ci->name;
+                    q0_dst->constr_x = 0;
+                    q0_dst->constr_y = 0;
+                    q0_dst->constr_z = -BelZ::iologic_0_z;
+                    q0_dst->constr_abs_z = false;
+                }
+
+                // disconnect Q0 output: it is wired internally
+                delete_nets.insert(ci->ports.at(id_Q0).net->name);
+                q0_dst->disconnectPort(id_I);
+                ci->disconnectPort(id_Q0);
+
+                ci->attrs[id_IOBUF] = 0;
+                // if Q1 is conected then disconnet it too
+                if (port_used(ci, id_Q1)) {
+                    q1_dst = net_only_drives(ctx, ci->ports.at(id_Q1).net, is_iob, id_OEN);
+                    if (q1_dst != nullptr) {
+                        delete_nets.insert(ci->ports.at(id_Q1).net->name);
+                        q0_dst->disconnectPort(id_OEN);
+                        ci->disconnectPort(id_Q1);
+                        ci->attrs[id_IOBUF] = 1;
+                    }
+                }
+                // if have XXX_ inputs connect them
+                if (ctx->ddr_has_extra_inputs) {
+                    ci->addInput(id_XXX_VSS);
+                    ci->connectPort(id_XXX_VSS, ctx->nets[ctx->id("$PACKER_GND_NET")].get());
+                    ci->addInput(id_XXX_VCC);
+                    ci->connectPort(id_XXX_VCC, ctx->nets[ctx->id("$PACKER_VCC_NET")].get());
+                }
+                if (ctx->gw1n9_quirk && iob_bel != q0_dst->attrs.end()) {
+                    bool have_XXX_VSS0 =
+                            ctx->bels[ctx->getBelByNameStr(iob_bel->second.as_string())].pins.count(id_XXX_VSS0);
+                    if (have_XXX_VSS0) {
+                        q0_dst->disconnectPort(id_XXX_VSS0);
+                        q0_dst->connectPort(id_XXX_VSS0, ctx->nets[ctx->id("$PACKER_VCC_NET")].get());
+                    }
+                }
+            } break;
+            default:
+                break;
+            }
+        }
+    }
+    for (auto pcell : packed_cells) {
+        ctx->cells.erase(pcell);
+    }
+    for (auto dnet : delete_nets) {
+        ctx->nets.erase(dnet);
+    }
+    for (auto &ncell : new_cells) {
+        ctx->cells[ncell->name] = std::move(ncell);
+    }
+}
+
+// Pack differential IO buffers
+static void pack_diff_io(Context *ctx)
+{
+    pool<IdString> packed_cells;
+    pool<IdString> delete_nets;
+
+    std::vector<std::unique_ptr<CellInfo>> new_cells;
+    log_info("Packing diff IOs..\n");
+
+    for (auto &cell : ctx->cells) {
+        CellInfo *ci = cell.second.get();
+        if (ctx->verbose)
+            log_info("cell '%s' is of type '%s'\n", ctx->nameOf(ci), ci->type.c_str(ctx));
+        if (is_gowin_diff_iob(ctx, ci)) {
+            CellInfo *iob_p = nullptr;
+            CellInfo *iob_n = nullptr;
+            switch (ci->type.index) {
+            case ID_TLVDS_OBUF: {
+                iob_p = net_only_drives(ctx, ci->ports.at(id_O).net, is_iob, id_I);
+                iob_n = net_only_drives(ctx, ci->ports.at(id_OB).net, is_iob, id_I);
+                NPNR_ASSERT(iob_p != nullptr);
+                NPNR_ASSERT(iob_n != nullptr);
+                auto iob_p_bel_a = iob_p->attrs.find(id_BEL);
+                if (iob_p_bel_a == iob_p->attrs.end()) {
+                    log_error("LVDS '%s' must be restricted.\n", ctx->nameOf(ci));
+                    continue;
+                }
+                BelId iob_p_bel = ctx->getBelByNameStr(iob_p_bel_a->second.as_string());
+                Loc loc_p = ctx->getBelLocation(iob_p_bel);
+                if (loc_p.z != 0) {
+                    log_error("LVDS '%s' positive pin is not A.\n", ctx->nameOf(ci));
+                    continue;
+                }
+                // restrict the N buffer
+                loc_p.z = 1;
+                iob_n->attrs[id_BEL] = ctx->getBelName(ctx->getBelByLocation(loc_p)).str(ctx);
+                // mark IOBs as part of DS pair
+                iob_n->attrs[id_DIFF] = std::string("N");
+                iob_n->attrs[id_DIFF_TYPE] = std::string("TLVDS_OBUF");
+                iob_p->attrs[id_DIFF] = std::string("P");
+                iob_p->attrs[id_DIFF_TYPE] = std::string("TLVDS_OBUF");
+                // disconnect N input: it is wired internally
+                delete_nets.insert(iob_n->ports.at(id_I).net->name);
+                iob_n->disconnectPort(id_I);
+                ci->disconnectPort(id_OB);
+                // disconnect P output
+                delete_nets.insert(ci->ports.at(id_O).net->name);
+                ci->disconnectPort(id_O);
+                // connect TLVDS input to P input
+                ci->movePortTo(id_I, iob_p, id_I);
+                packed_cells.insert(ci->name);
+            } break;
+            default:
+                break;
+            }
+        }
+    }
+    for (auto pcell : packed_cells) {
+        ctx->cells.erase(pcell);
+    }
+    for (auto dnet : delete_nets) {
+        ctx->nets.erase(dnet);
+    }
+    for (auto &ncell : new_cells) {
+        ctx->cells[ncell->name] = std::move(ncell);
+    }
+}
+
 // Pack IO buffers
 static void pack_io(Context *ctx)
 {
@@ -688,6 +909,8 @@ static void pack_io(Context *ctx)
 
     for (auto &cell : ctx->cells) {
         CellInfo *ci = cell.second.get();
+        if (ctx->verbose)
+            log_info("cell '%s' is of type '%s'\n", ctx->nameOf(ci), ci->type.c_str(ctx));
         if (is_gowin_iob(ctx, ci)) {
             CellInfo *iob = nullptr;
             switch (ci->type.index) {
@@ -710,16 +933,46 @@ static void pack_io(Context *ctx)
                 // delete the $nexpnr_[io]buf
                 for (auto &p : iob->ports) {
                     IdString netname = p.second.net->name;
-                    disconnect_port(ctx, iob, p.first);
+                    iob->disconnectPort(p.first);
                     delete_nets.insert(netname);
                 }
                 packed_cells.insert(iob->name);
             }
+            // what type to create
+            IdString new_cell_type = id_IOB;
+            std::string constr_bel_name = std::string("");
+            bool have_xxx_port = false;
+            // check whether the given IO is limited to simplified IO cells
+            auto constr_bel = ci->attrs.find(id_BEL);
+            if (constr_bel != ci->attrs.end()) {
+                constr_bel_name = constr_bel->second.as_string();
+            }
+            constr_bel = iob->attrs.find(id_BEL);
+            if (constr_bel != iob->attrs.end()) {
+                constr_bel_name = constr_bel->second.as_string();
+            }
+            if (!constr_bel_name.empty()) {
+                BelId constr_bel = ctx->getBelByNameStr(constr_bel_name);
+                if (constr_bel != BelId()) {
+                    new_cell_type = ctx->bels[constr_bel].type;
+                    if (ctx->gw1n9_quirk) {
+                        have_xxx_port = ctx->bels[constr_bel].pins.count(id_XXX_VSS0) != 0;
+                    }
+                }
+            }
+
             // Create a IOB buffer
-            std::unique_ptr<CellInfo> ice_cell = create_generic_cell(ctx, id_IOB, ci->name.str(ctx) + "$iob");
+            std::unique_ptr<CellInfo> ice_cell = create_generic_cell(ctx, new_cell_type, ci->name.str(ctx) + "$iob");
             gwio_to_iob(ctx, ci, ice_cell.get(), packed_cells);
             new_cells.push_back(std::move(ice_cell));
             auto gwiob = new_cells.back().get();
+            // XXX GW1NR-9 quirks
+            if (have_xxx_port && ci->type != id_IBUF) {
+                gwiob->addInput(id_XXX_VSS0);
+                gwiob->connectPort(id_XXX_VSS0, ctx->nets[ctx->id("$PACKER_GND_NET")].get());
+                gwiob->addInput(id_XXX_VSS1);
+                gwiob->connectPort(id_XXX_VSS1, ctx->nets[ctx->id("$PACKER_GND_NET")].get());
+            }
 
             packed_cells.insert(ci->name);
             if (iob != nullptr) {
@@ -755,12 +1008,15 @@ bool Arch::pack()
     try {
         log_break();
         pack_constants(ctx);
+        pack_gsr(ctx);
         pack_io(ctx);
+        pack_diff_io(ctx);
+        pack_iologic(ctx);
         pack_wideluts(ctx);
         pack_alus(ctx);
         pack_lut_lutffs(ctx);
         pack_nonlut_ffs(ctx);
-        ctx->settings[ctx->id("pack")] = 1;
+        ctx->settings[id_pack] = 1;
         ctx->assignArchInfo();
         log_info("Checksum: 0x%08x\n", ctx->checksum());
         return true;
