@@ -41,13 +41,13 @@ struct MistralPacker
         vcc_drv->addOutput(id_Q);
         gnd_net = ctx->createNet(ctx->id("$PACKER_GND_NET"));
         vcc_net = ctx->createNet(ctx->id("$PACKER_VCC_NET"));
-        connect_port(ctx, gnd_net, gnd_drv, id_Q);
-        connect_port(ctx, vcc_net, vcc_drv, id_Q);
+        gnd_drv->connectPort(id_Q, gnd_net);
+        vcc_drv->connectPort(id_Q, vcc_net);
     }
 
     CellPinState get_pin_needed_muxval(CellInfo *cell, IdString port)
     {
-        NetInfo *net = get_net_or_empty(cell, port);
+        NetInfo *net = cell->getPort(port);
         if (net == nullptr || net->driver.cell == nullptr) {
             // Pin is disconnected
             // If a mux value exists already, honour it
@@ -78,14 +78,14 @@ struct MistralPacker
     void uninvert_port(CellInfo *cell, IdString port)
     {
         // Rewire a port so it is driven by the input to an inverter
-        NetInfo *net = get_net_or_empty(cell, port);
+        NetInfo *net = cell->getPort(port);
         NPNR_ASSERT(net != nullptr && net->driver.cell != nullptr && net->driver.cell->type == id_MISTRAL_NOT);
         CellInfo *inv = net->driver.cell;
-        disconnect_port(ctx, cell, port);
+        cell->disconnectPort(port);
 
-        NetInfo *inv_a = get_net_or_empty(inv, id_A);
+        NetInfo *inv_a = inv->getPort(id_A);
         if (inv_a != nullptr) {
-            connect_port(ctx, inv_a, cell, port);
+            cell->connectPort(port, inv_a);
         }
     }
 
@@ -117,12 +117,12 @@ struct MistralPacker
                 // Pin is tied to a constant
                 // If there is a hard constant option; use it
                 if ((pin_style & int(req_mux)) == req_mux) {
-                    disconnect_port(ctx, cell, port_name);
+                    cell->disconnectPort(port_name);
                     cell->pin_data[port_name].state = req_mux;
                 } else {
-                    disconnect_port(ctx, cell, port_name);
+                    cell->disconnectPort(port_name);
                     // There is no hard constant, we need to connect it to the relevant soft-constant net
-                    connect_port(ctx, (req_mux == PIN_1) ? vcc_net : gnd_net, cell, port_name);
+                    cell->connectPort(port_name, (req_mux == PIN_1) ? vcc_net : gnd_net);
                 }
             }
         }
@@ -138,7 +138,7 @@ struct MistralPacker
             if (ci->type != id_MISTRAL_NOT && ci->type != id_GND && ci->type != id_VCC)
                 continue;
             IdString port = (ci->type == id_MISTRAL_NOT) ? id_Q : id_Y;
-            NetInfo *out = get_net_or_empty(ci, port);
+            NetInfo *out = ci->getPort(port);
             if (out == nullptr) {
                 trim_cells.push_back(ci->name);
                 continue;
@@ -146,7 +146,7 @@ struct MistralPacker
             if (!out->users.empty())
                 continue;
 
-            disconnect_port(ctx, ci, id_A);
+            ci->disconnectPort(id_A);
 
             trim_cells.push_back(ci->name);
             trim_nets.push_back(out->name);
@@ -174,7 +174,7 @@ struct MistralPacker
                 continue;
             if (ci->get_pin_state(id_SLOAD) != PIN_0)
                 continue;
-            disconnect_port(ctx, ci, id_SDATA);
+            ci->disconnectPort(id_SDATA);
         }
         // Remove superfluous inverters and constant drivers
         trim_design();
@@ -197,27 +197,27 @@ struct MistralPacker
             if (ci->type == ctx->id("$nextpnr_ibuf") || ci->type == ctx->id("$nextpnr_iobuf")) {
                 // Might have an input buffer (IB etc) connected to it
                 is_npnr_iob = true;
-                NetInfo *o = get_net_or_empty(ci, id_O);
+                NetInfo *o = ci->getPort(id_O);
                 if (o == nullptr)
                     ;
-                else if (o->users.size() > 1)
+                else if (o->users.entries() > 1)
                     log_error("Top level pin '%s' has multiple input buffers\n", ctx->nameOf(port.first));
-                else if (o->users.size() == 1)
-                    top_port = o->users.at(0);
+                else if (o->users.entries() == 1)
+                    top_port = *o->users.begin();
             }
             if (ci->type == ctx->id("$nextpnr_obuf") || ci->type == ctx->id("$nextpnr_iobuf")) {
                 // Might have an output buffer (OB etc) connected to it
                 is_npnr_iob = true;
-                NetInfo *i = get_net_or_empty(ci, id_I);
+                NetInfo *i = ci->getPort(id_I);
                 if (i != nullptr && i->driver.cell != nullptr) {
                     if (top_port.cell != nullptr)
                         log_error("Top level pin '%s' has multiple input/output buffers\n", ctx->nameOf(port.first));
                     top_port = i->driver;
                 }
                 // Edge case of a bidirectional buffer driving an output pin
-                if (i->users.size() > 2) {
+                if (i->users.entries() > 2) {
                     log_error("Top level pin '%s' has illegal buffer configuration\n", ctx->nameOf(port.first));
-                } else if (i->users.size() == 2) {
+                } else if (i->users.entries() == 2) {
                     if (top_port.cell != nullptr)
                         log_error("Top level pin '%s' has illegal buffer configuration\n", ctx->nameOf(port.first));
                     for (auto &usr : i->users) {
@@ -245,8 +245,8 @@ struct MistralPacker
                 port.second.net = top_port.cell->ports.at(top_port.port).net;
             }
             // Now remove the nextpnr-inserted buffer
-            disconnect_port(ctx, ci, id_I);
-            disconnect_port(ctx, ci, id_O);
+            ci->disconnectPort(id_I);
+            ci->disconnectPort(id_O);
             ctx->cells.erase(port.first);
         }
     }
@@ -290,19 +290,19 @@ struct MistralPacker
             CellInfo *ci = cell.second.get();
             if (ci->type != id_MISTRAL_ALUT_ARITH)
                 continue;
-            const NetInfo *cin = get_net_or_empty(ci, id_CI);
+            const NetInfo *cin = ci->getPort(id_CI);
             if (cin != nullptr && cin->driver.cell != nullptr)
                 continue; // not the start of a chain
             std::vector<CellInfo *> chain;
             CellInfo *cursor = ci;
             while (true) {
                 chain.push_back(cursor);
-                const NetInfo *co = get_net_or_empty(cursor, id_CO);
+                const NetInfo *co = cursor->getPort(id_CO);
                 if (co == nullptr || co->users.empty())
                     break;
-                if (co->users.size() > 1)
+                if (co->users.entries() > 1)
                     log_error("Carry net %s has more than one sink!\n", ctx->nameOf(co));
-                auto &usr = co->users.at(0);
+                auto &usr = *co->users.begin();
                 if (usr.port != id_CI)
                     log_error("Carry net %s drives port %s, expected CI\n", ctx->nameOf(co), ctx->nameOf(usr.port));
                 cursor = usr.cell;
@@ -327,7 +327,7 @@ struct MistralPacker
                 for (int i = 0; i < int(chain.size()); i++) {
                     auto &c = chain.at(i);
                     log_info("    i=%d cell=%s dy=%d z=%d ci=%s co=%s\n", i, ctx->nameOf(c), c->constr_y, c->constr_z,
-                             ctx->nameOf(get_net_or_empty(c, id_CI)), ctx->nameOf(get_net_or_empty(c, id_CO)));
+                             ctx->nameOf(c->getPort(id_CI)), ctx->nameOf(c->getPort(id_CO)));
                 }
             }
         }
@@ -338,7 +338,7 @@ struct MistralPacker
                 continue;
             if (ci->cluster == ClusterId())
                 log_error("Failed to include arith cell '%s' in any chain (CI=%s)\n", ctx->nameOf(ci),
-                          ctx->nameOf(get_net_or_empty(ci, id_CI)));
+                          ctx->nameOf(ci->getPort(id_CI)));
         }
     }
 
@@ -382,6 +382,86 @@ struct MistralPacker
         }
     }
 
+    void setup_m10ks()
+    {
+        for (auto &cell : ctx->cells) {
+            CellInfo *ci = cell.second.get();
+            if (ci->type != id_MISTRAL_M10K)
+                continue;
+
+            auto abits = ci->params.at(id_CFG_ABITS).as_int64();
+            auto dbits = ci->params.at(id_CFG_DBITS).as_int64();
+            NPNR_ASSERT(abits >= 7 && abits <= 13);
+            NPNR_ASSERT(dbits == 1 || dbits == 2 || dbits == 5 || dbits == 10 || dbits == 20);
+
+            // Quartus doesn't seem to generate ADDRSTALL[AB], BYTEENABLE[AB][01].
+
+            // It *does* generate ACLR[01] but leaves them unconnected if unused.
+
+            // Enables.
+            // RDEN[1] and WREN[0] are left unconnected.
+            ci->pin_data[ctx->id("A1EN")].bel_pins = {ctx->id("WREN[1]")};
+            ci->pin_data[ctx->id("B1EN")].bel_pins = {ctx->id("RDEN[0]")};
+
+            // Clocks.
+            ci->pin_data[ctx->id("CLK1")].bel_pins = {ctx->id("CLKIN[0]")};
+
+            // Enables left unconnected.
+
+            // Address lines.
+            int addr_offset = std::max(12 - std::max(abits, int64_t{9}), 0l);
+            int bit_offset = 0;
+            if (abits == 13) {
+                ci->pin_data[ctx->id("A1ADDR[0]")].bel_pins = {ctx->id("DATAAIN[4]")};
+                ci->pin_data[ctx->id("B1ADDR[0]")].bel_pins = {ctx->id("DATABIN[19]")};
+                bit_offset = 1;
+            }
+            for (int bit = bit_offset; bit < abits; bit++) {
+                ci->pin_data[ctx->id(stringf("A1ADDR[%d]", bit))].bel_pins = {
+                        ctx->id(stringf("ADDRA[%d]", bit + addr_offset))};
+                ci->pin_data[ctx->id(stringf("B1ADDR[%d]", bit))].bel_pins = {
+                        ctx->id(stringf("ADDRB[%d]", bit + addr_offset))};
+            }
+
+            // Data lines
+            std::vector<int> offsets;
+            offsets.push_back(0);
+            if (abits >= 10 && dbits <= 10) {
+                offsets.push_back(10);
+            }
+            if (abits >= 11 && dbits <= 5) {
+                offsets.push_back(5);
+                offsets.push_back(15);
+            }
+            if (abits >= 12 && dbits <= 2) {
+                offsets.push_back(2);
+                offsets.push_back(7);
+                offsets.push_back(12);
+                offsets.push_back(17);
+            }
+            if (abits == 13 && dbits == 1) {
+                offsets.push_back(1);
+                offsets.push_back(3);
+                offsets.push_back(6);
+                offsets.push_back(8);
+                offsets.push_back(11);
+                offsets.push_back(13);
+                offsets.push_back(16);
+                offsets.push_back(18);
+            }
+            for (int bit = 0; bit < dbits; bit++) {
+                for (int offset : offsets) {
+                    ci->pin_data[ctx->id(stringf("A1DATA[%d]", bit))].bel_pins.push_back(
+                            ctx->id(stringf("DATAAIN[%d]", bit + offset)));
+                }
+            }
+
+            for (int bit = 0; bit < dbits; bit++) {
+                ci->pin_data[ctx->id(stringf("B1DATA[%d]", bit))].bel_pins = {ctx->id(stringf("DATABOUT[%d]", bit))};
+            }
+        }
+    }
+
     void run()
     {
         init_constant_nets();
@@ -389,6 +469,7 @@ struct MistralPacker
         pack_io();
         constrain_carries();
         constrain_lutram();
+        setup_m10ks();
     }
 };
 }; // namespace
