@@ -119,6 +119,9 @@ Arch::Arch(ArchArgs args) : args(args)
             ts.bels_by_z[bel.z].tile = i;
             ts.bels_by_z[bel.z].index = j;
         }
+        auto &ts = tileStatus.at(i);
+        ts.boundwires.resize(loc.wires.size());
+        ts.boundpips.resize(loc.pips.size());
     }
 
     for (int i = 0; i < chip_info->width; i++) {
@@ -182,7 +185,12 @@ Arch::Arch(ArchArgs args) : args(args)
             for (auto pip : getPipsUphill(w))
                 disabled_pips.insert(pip);
         }
-        NPNR_ASSERT(disabled_pips.size() == 4);
+        // TODO: find a better solution to disable these
+        WireId dcs_out =
+                getWireByName(IdStringList(std::array<IdString, 3>{x_ids.at(37), y_ids.at(10), id_JDCSOUT_DCS_DCSIP}));
+        for (auto dcs_pip : getPipsUphill(dcs_out))
+            disabled_pips.insert(dcs_pip);
+        NPNR_ASSERT(disabled_pips.size() == 6);
     }
 }
 
@@ -292,13 +300,13 @@ std::vector<std::pair<IdString, std::string>> Arch::getBelAttrs(BelId bel) const
 {
     std::vector<std::pair<IdString, std::string>> ret;
 
-    ret.emplace_back(id("INDEX"), stringf("%d", bel.index));
+    ret.emplace_back(id_INDEX, stringf("%d", bel.index));
 
-    ret.emplace_back(id("GRID_X"), stringf("%d", bel.tile % chip_info->width));
-    ret.emplace_back(id("GRID_Y"), stringf("%d", bel.tile / chip_info->width));
-    ret.emplace_back(id("BEL_Z"), stringf("%d", bel_data(bel).z));
+    ret.emplace_back(id_GRID_X, stringf("%d", bel.tile % chip_info->width));
+    ret.emplace_back(id_GRID_Y, stringf("%d", bel.tile / chip_info->width));
+    ret.emplace_back(id_BEL_Z, stringf("%d", bel_data(bel).z));
 
-    ret.emplace_back(id("BEL_TYPE"), nameOf(getBelType(bel)));
+    ret.emplace_back(id_BEL_TYPE, nameOf(getBelType(bel)));
 
     return ret;
 }
@@ -329,11 +337,11 @@ std::vector<std::pair<IdString, std::string>> Arch::getWireAttrs(WireId wire) co
 {
     std::vector<std::pair<IdString, std::string>> ret;
 
-    ret.emplace_back(id("INDEX"), stringf("%d", wire.index));
+    ret.emplace_back(id_INDEX, stringf("%d", wire.index));
 
-    ret.emplace_back(id("GRID_X"), stringf("%d", wire.tile % chip_info->width));
-    ret.emplace_back(id("GRID_Y"), stringf("%d", wire.tile / chip_info->width));
-    ret.emplace_back(id("FLAGS"), stringf("%u", wire_data(wire).flags));
+    ret.emplace_back(id_GRID_X, stringf("%d", wire.tile % chip_info->width));
+    ret.emplace_back(id_GRID_Y, stringf("%d", wire.tile / chip_info->width));
+    ret.emplace_back(id_FLAGS, stringf("%u", wire_data(wire).flags));
 
     return ret;
 }
@@ -380,13 +388,13 @@ std::vector<std::pair<IdString, std::string>> Arch::getPipAttrs(PipId pip) const
 {
     std::vector<std::pair<IdString, std::string>> ret;
 
-    ret.emplace_back(id("INDEX"), stringf("%d", pip.index));
+    ret.emplace_back(id_INDEX, stringf("%d", pip.index));
 
-    ret.emplace_back(id("GRID_X"), stringf("%d", pip.tile % chip_info->width));
-    ret.emplace_back(id("GRID_Y"), stringf("%d", pip.tile / chip_info->width));
+    ret.emplace_back(id_GRID_X, stringf("%d", pip.tile % chip_info->width));
+    ret.emplace_back(id_GRID_Y, stringf("%d", pip.tile / chip_info->width));
 
-    ret.emplace_back(id("FROM_TILE_WIRE"), nameOf(IdString(loc_data(pip).wires[pip_data(pip).from_wire].name)));
-    ret.emplace_back(id("TO_TILE_WIRE"), nameOf(IdString(loc_data(pip).wires[pip_data(pip).to_wire].name)));
+    ret.emplace_back(id_FROM_TILE_WIRE, nameOf(IdString(loc_data(pip).wires[pip_data(pip).from_wire].name)));
+    ret.emplace_back(id_TO_TILE_WIRE, nameOf(IdString(loc_data(pip).wires[pip_data(pip).to_wire].name)));
 
     return ret;
 }
@@ -478,6 +486,12 @@ bool Arch::getCellDelay(const CellInfo *cell, IdString fromPort, IdString toPort
         if (fromPort == id_CLK)
             return false; // don't include delays that are actually clock-to-out here
         return lookup_cell_delay(cell->tmg_index, lookup_port(fromPort), lookup_port(toPort), delay);
+    } else if (cell->type == id_DCS) {
+        if (fromPort == id_SELFORCE || fromPort == id_SEL) {
+            return false;
+        }
+        int index = get_cell_timing_idx(id_DCS, id_DCS);
+        return lookup_cell_delay(index, fromPort, toPort, delay);
     }
     return false;
 }
@@ -538,6 +552,23 @@ TimingPortClass Arch::getPortTimingClass(const CellInfo *cell, IdString port, in
         if (type == TMG_REGISTER_INPUT || type == TMG_REGISTER_OUTPUT)
             clockInfoCount = 1;
         return type;
+    } else if (cell->type == id_DCC) {
+        if (port == id_CLKI)
+            return TMG_CLOCK_INPUT;
+        else if (port == id_CLKO)
+            return TMG_GEN_CLOCK;
+        else if (port == id_CE)
+            return TMG_COMB_INPUT;
+    } else if (cell->type == id_DCS) {
+        // FIXME: Making inputs TMG_CLOCK_INPUT and the output TMG_CLOCK_GEN
+        // yielded in error in the timing analyzer. For now keep those as
+        // regular combinational ports.
+        if (port == id_CLK0 || port == id_CLK1)
+            return TMG_COMB_INPUT;
+        else if (port == id_DCSOUT) {
+            return TMG_COMB_OUTPUT;
+        }
+        return TMG_IGNORE;
     }
     return TMG_IGNORE;
 }
@@ -598,18 +629,17 @@ delay_t Arch::estimateDelay(WireId src, WireId dst) const
     int dst_x = dst.tile % chip_info->width, dst_y = dst.tile / chip_info->width;
     int dist_x = std::abs(src_x - dst_x);
     int dist_y = std::abs(src_y - dst_y);
-    return 75 * dist_x + 75 * dist_y + 250;
-}
-delay_t Arch::predictDelay(const NetInfo *net_info, const PortRef &sink) const
-{
-    if (net_info->driver.cell == nullptr || net_info->driver.cell->bel == BelId() || sink.cell->bel == BelId())
-        return 0;
-    if (sink.port == id_FCI)
-        return 0;
-    int src_x = net_info->driver.cell->bel.tile % chip_info->width,
-        src_y = net_info->driver.cell->bel.tile / chip_info->width;
 
-    int dst_x = sink.cell->bel.tile % chip_info->width, dst_y = sink.cell->bel.tile / chip_info->width;
+    return estimate_delay_mult * (dist_x + dist_y) + 250;
+}
+delay_t Arch::predictDelay(BelId src_bel, IdString src_pin, BelId dst_bel, IdString dst_pin) const
+{
+    NPNR_UNUSED(src_pin);
+    if (dst_pin == id_FCI)
+        return 0;
+    int src_x = src_bel.tile % chip_info->width, src_y = src_bel.tile / chip_info->width;
+
+    int dst_x = dst_bel.tile % chip_info->width, dst_y = dst_bel.tile / chip_info->width;
     int dist_x = std::abs(src_x - dst_x);
     int dist_y = std::abs(src_y - dst_y);
     return 100 * dist_x + 100 * dist_y + 250;
@@ -654,7 +684,11 @@ ArcBounds Arch::getRouteBoundingBox(WireId src, WireId dst) const
 
 bool Arch::place()
 {
-    std::string placer = str_or_default(settings, id("placer"), defaultPlacer);
+    estimate_delay_mult = 75;
+    if (getCtx()->settings.count(getCtx()->id("estimate-delay-mult")))
+        estimate_delay_mult = getCtx()->setting<int>("estimate-delay-mult");
+
+    std::string placer = str_or_default(settings, id_placer, defaultPlacer);
 
     if (placer == "heap") {
         PlacerHeapCfg cfg(getCtx());
@@ -678,7 +712,7 @@ bool Arch::place()
 
     post_place_opt();
 
-    getCtx()->attrs[getCtx()->id("step")] = std::string("place");
+    getCtx()->attrs[id_step] = std::string("place");
     archInfoToAttributes();
     return true;
 }
@@ -729,7 +763,7 @@ bool Arch::route()
 
     route_globals();
 
-    std::string router = str_or_default(settings, id("router"), defaultRouter);
+    std::string router = str_or_default(settings, id_router, defaultRouter);
     bool result;
     if (router == "router1") {
         result = router1(getCtx(), Router1Cfg(getCtx()));
@@ -741,7 +775,7 @@ bool Arch::route()
     } else {
         log_error("Nexus architecture does not support router '%s'\n", router.c_str());
     }
-    getCtx()->attrs[getCtx()->id("step")] = std::string("route");
+    getCtx()->attrs[id_step] = std::string("route");
     archInfoToAttributes();
     return result;
 }
